// MNG browser plugin
// By Jason Summers
// Based on libmng by Gerard Juyn


#include <windows.h>
#include <commdlg.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <malloc.h>
#include <stdarg.h>


#include <jversion.h>      // part of libjpeg
#include <zlib.h>          // for zlibVersion

#ifdef MNGPLG
#include "resourcepl.h"
#define MNGCTRL_NAME "MNGPLG"
#endif

#ifdef MNG4IE
#include "resourceac.h"
#define MNGCTRL_NAME "MNG4IE"
#endif

#ifdef MNGSV
#include "resourcesv.h"
#define MNGCTRL_NAME "MNGSV"
#endif

#include "mngctrl.h"

#ifndef LCMS_VERSION
#define LCMS_VERSION 0
#endif

//#include "npapidefs.h"

#define MNGCTRL_CMS
//#define MNGCTRL_TRACE

#define IDBASE           47000
#define ID_SAVEAS        (IDBASE+0)
#define ID_COPYIMAGE     (IDBASE+1)
#define ID_COPYURL       (IDBASE+2)
#define ID_VIEWIMAGE     (IDBASE+3)
#define ID_ABOUT         (IDBASE+4)
#define ID_FREEZE        (IDBASE+5)
#define ID_RESTARTANIM   (IDBASE+6)
#define ID_COPYLINKLOC   (IDBASE+7)
#define ID_STOPANIM      (IDBASE+8)
#define ID_SHOWERROR     (IDBASE+9)
#define ID_PROPERTIES    (IDBASE+10)


/* global variables */
struct globals_struct globals;

#ifdef MNGCTRL_TRACE
static FILE *tracefile;
#endif

static const char* gInstanceLookupString = "pdata";

/* function prototypes */
static LRESULT CALLBACK DlgProcAbout(HWND hWnd, UINT Msg, WPARAM wParam, LPARAM lParam);
static LRESULT CALLBACK DlgProcProp(HWND hWnd, UINT Msg, WPARAM wParam, LPARAM lParam);
static LRESULT CALLBACK PluginWindowProc(HWND hWnd, UINT Msg, WPARAM wParam, LPARAM lParam);

static void set_scrollbars(PluginInstance *This);


//////////////////////////// NPN_functions



/* ----------------------------------------------------------------------- */

static void warn(PluginInstance *This, char *fmt, ...)
{
	va_list ap;
	char buf[2048];
	HWND hwnd;

	va_start(ap, fmt);
	wvsprintf(buf,fmt, ap);
	va_end(ap);

	if(This) hwnd= This->fhWnd;
	else hwnd=NULL;

	MessageBox(hwnd,buf,globals.appname,MB_OK|MB_ICONWARNING);
}


static void set_error(PluginInstance *This, char *fmt, ...)
{
	va_list ap;
	char buf[2048];
	HWND hwnd;

	va_start(ap, fmt);
	wvsprintf(buf,fmt, ap);
	va_end(ap);

	if(This) hwnd= This->fhWnd;
	else hwnd=NULL;

	This->errorflag=1;
	lstrcpyn(This->errormsg,buf,256);

	if(This->lpdib) {
		free(This->lpdib);
		This->lpdib=NULL;
	}
	This->xscrollpos = This->yscrollpos = 0;

	if(This->fhWnd)
		InvalidateRect(This->fhWnd,NULL,TRUE);
}

/* ----------------------------------------------------------------------- */
//    MNG callbacks
#define MNGCTRL_CALLBACK MNG_DECL

static mng_ptr MNGCTRL_CALLBACK memallocfunc(mng_size_t n)
{
	return (mng_ptr) calloc(n,1);
}


static void MNGCTRL_CALLBACK memfreefunc(mng_ptr p, mng_size_t n)
{
	free((void*)p);
}


static mng_bool MNGCTRL_CALLBACK callback_openstream (mng_handle mng)
{
//	PluginInstance *This;
//	This = (PluginInstance*) mng_get_userdata(mng);
	return MNG_TRUE;
}

static mng_bool MNGCTRL_CALLBACK callback_closestream (mng_handle mng)
{
	PluginInstance *This;
	This = (PluginInstance*) mng_get_userdata(mng);
	This->loadstate = STATE_LOADED;  // this is probably redundant

	return MNG_TRUE;
}


static mng_bool MNGCTRL_CALLBACK callback_readdata (mng_handle mng,mng_ptr pBuf,
                      mng_uint32 Buflen,mng_uint32 *pRead)
{
	int n;
	PluginInstance *This;
	This = (PluginInstance*) mng_get_userdata(mng);

#ifdef MNGCTRL_TRACE
	fprintf(tracefile,"readdata callback buflen=%d loadstate=%d bytesloaded=%d libmngpos=%d\n",
		Buflen,This->loadstate,This->bytesloaded, This->libmngpos);
#endif


	// do we have enough data available?
	if(This->bytesloaded - This->libmngpos >= Buflen) {
		CopyMemory(pBuf,&This->mngdata[This->libmngpos],Buflen);
		(*pRead)= Buflen;
		This->libmngpos += Buflen;

#ifdef MNGCTRL_TRACE
	fprintf(tracefile,"returning full: %d\n",Buflen);
#endif
		This->byteswanted=0;
		return MNG_TRUE;
	}
	else if(This->loadstate>=STATE_LOADED) {
		// We don't have the data it asked for, but we're at the end
		// of file, so send it anyway...?

		n=This->bytesloaded-This->libmngpos;

		if(n>0) {
			CopyMemory(pBuf,&This->mngdata[This->libmngpos],n);
			This->libmngpos+=n;
		}
		(*pRead)=n;
		// so what do we return?
#ifdef MNGCTRL_TRACE
	fprintf(tracefile,"returning partial: %d\n",n);
#endif
		This->byteswanted=0;
		return MNG_TRUE;
	}

	// else we don't yet have the data it's requesting
#ifdef MNGCTRL_TRACE
	fprintf(tracefile,"returning 0\n");
#endif
	(*pRead)=0;
	This->byteswanted=Buflen;
	return MNG_TRUE;
}


static mng_bool MNGCTRL_CALLBACK callback_processheader(mng_handle mng,mng_uint32 iWidth,mng_uint32 iHeight)
{
	PluginInstance *This;
	This = (PluginInstance*) mng_get_userdata(mng);

	This->diblinesize = (((iWidth * 24)+31)/32)*4;
	This->dibsize = sizeof(BITMAPINFOHEADER) + This->diblinesize*iHeight;
	This->lpdib = calloc(This->dibsize,1);
	This->lpdibinfo = (LPBITMAPINFOHEADER)This->lpdib;
	This->lpdibbits = &This->lpdib[sizeof(BITMAPINFOHEADER)];
	ZeroMemory((void*)This->lpdib,sizeof(BITMAPINFOHEADER));
	This->lpdibinfo->biSize = sizeof(BITMAPINFOHEADER);
	This->lpdibinfo->biWidth = iWidth;
	This->lpdibinfo->biHeight = iHeight;
	This->lpdibinfo->biPlanes = 1;
	This->lpdibinfo->biBitCount = 24;

	mng_set_canvasstyle (mng, MNG_CANVAS_BGR8);

/*	if(This->fhWnd) {
		if((int)iWidth > This->windowwidth || (int)iHeight > This->windowheight) {
			This->scrolling=1;
		}
	} */

	set_scrollbars(This);
	return MNG_TRUE;
}


static mng_ptr MNGCTRL_CALLBACK callback_getcanvasline (mng_handle mng, mng_uint32 iLinenr)
{
	unsigned char *pp;

	PluginInstance *This;
	This = (PluginInstance*) mng_get_userdata(mng);
	pp = (&This->lpdibbits[(This->lpdibinfo->biHeight-1-iLinenr)*This->diblinesize]);
	return (mng_ptr) pp;
}

static mng_bool MNGCTRL_CALLBACK callback_refresh (mng_handle mng, mng_uint32 iLeft, mng_uint32 iTop,
                       mng_uint32 iRight, mng_uint32 iBottom)
{
	PluginInstance *This;
	RECT rect;

	This = (PluginInstance*) mng_get_userdata(mng);

	if(This->loadstate<STATE_VALIDFRAME) {
		This->loadstate=STATE_VALIDFRAME;
	}

	if(This->fhWnd) {
		if(This->paintedyet) {
			rect.left= iLeft      - This->xscrollpos;
			rect.top= iTop        - This->yscrollpos;
			rect.right= iLeft+iRight;
			rect.bottom= iTop+iBottom;

			InvalidateRect(This->fhWnd,&rect,FALSE);
		}
		else {
			// Make sure the first paint clears the whole plugin window
			InvalidateRect(This->fhWnd,NULL,TRUE);
			This->paintedyet=1;
		}
		UpdateWindow(This->fhWnd);
	}
	return MNG_TRUE;
}

static mng_uint32 MNGCTRL_CALLBACK callback_gettickcount (mng_handle mng)
{
	return GetTickCount();
}


static mng_bool MNGCTRL_CALLBACK callback_settimer (mng_handle mng,mng_uint32 iMsecs)
{
	PluginInstance *This;
	This = (PluginInstance*) mng_get_userdata(mng);

	if(This->fhWnd) {
		if(!SetTimer(This->fhWnd,1,(UINT)iMsecs,NULL)) {
			warn(This,"Unable to create a timer for animation");
			This->frozen=1;
			//return MNG_FALSE;
			return MNG_TRUE;
		}
		This->timer_set=1;
	}
	return MNG_TRUE;
}

static mng_bool MNGCTRL_CALLBACK callback_processtext(mng_handle mng,
    mng_uint8 iType, mng_pchar zKeyword, mng_pchar zText,
    mng_pchar zLanguage, mng_pchar zTranslation)
{
	PluginInstance *This;
	int pos,i;

	This = (PluginInstance*) mng_get_userdata(mng);

	if(!This->textdata) {
		This->textdata=(char*)malloc(MAXLEN_TEXT+10);
		if(!This->textdata) return MNG_TRUE;
		lstrcpy(This->textdata,"");
	}

	pos=lstrlen(This->textdata);
	if(pos>=(MAXLEN_TEXT-10)) return MNG_TRUE;

	if(pos>0) {    /* separate items with a blank line */
		This->textdata[pos++]='\r';
		This->textdata[pos++]='\n';
		This->textdata[pos++]='\r';
		This->textdata[pos++]='\n';
	}

	for(i=0;zKeyword[i];i++) {
		if(pos<MAXLEN_TEXT)
			This->textdata[pos++]=zKeyword[i];
	}
	This->textdata[pos++]=':';
	This->textdata[pos++]=' ';

	for(i=0;zText[i];i++) {
		if(pos<MAXLEN_TEXT) {
			if(zText[i]=='\n') {
				This->textdata[pos++]='\r';
			}
			This->textdata[pos++]=zText[i];
		}
	}
	This->textdata[pos++]='\0';

	return MNG_TRUE;
}

#ifdef MNGCTRL_TRACE
static mng_bool MNGCTRL_CALLBACK callback_traceproc (mng_handle mng,
	mng_int32  iFuncnr,
	mng_int32  iFuncseq,
	mng_pchar  zFuncname)
{
	if(tracefile) {
		fprintf(tracefile,"%d\t%d\t%d\t%s\n",(int)mng,iFuncnr,iFuncseq,zFuncname);
	}
	return MNG_TRUE;
}
#endif

/* ----------------------------------------------------------------------- */
#if 0
static int file_exists(const char *fn)
{
   HANDLE h;

   // try to open with no access
   h=CreateFile(fn,0,FILE_SHARE_READ,NULL,OPEN_EXISTING,
      FILE_ATTRIBUTE_NORMAL,NULL);
   if(h == INVALID_HANDLE_VALUE) { return 0; }
   CloseHandle(h);
   return 1;
}
#endif

static void handle_read_error(PluginInstance *This, mng_retcode rv)
{
	mng_int8     iSeverity;
	mng_chunkid  iChunkname;
	mng_uint32   iChunkseq;
	mng_int32    iExtra1;
	mng_int32    iExtra2;
	mng_pchar    zErrortext;

#ifdef MNGCTRL_TRACE
	fprintf(tracefile,"returned: %d\n",rv);
#endif

	switch(rv) {
	case MNG_NOERROR: case MNG_NEEDTIMERWAIT:
		break;

	case MNG_NEEDMOREDATA:
		if(This->loadstate>=STATE_LOADED) {
			set_error(This,"Unexpected end of file");
		}
		else {
			This->needresume=1;
		}
		break;

	case MNG_INVALIDSIG:
		if(globals.arch==PLGARCH_STANDALONE)
			set_error(This,"Invalid MNG file");
		else
			set_error(This,"Invalid or missing MNG file (maybe a 404 Not Found error)");
		break;

	default:

		mng_getlasterror(This->mng, &iSeverity,&iChunkname,&iChunkseq,&iExtra1,
			&iExtra2,&zErrortext);

		if(zErrortext) {
			set_error(This,"Error reported by libmng (%d)\r\n\r\n%s",(int)rv,zErrortext);
		}
		else {
			set_error(This,"Error %d reported by libmng",(int)rv);
		}
	}
}

#ifdef MNGCTRL_CMS

static int init_color_management(PluginInstance *This)
{
	mng_set_outputsrgb(This->mng);
	return 1;
}

#endif

// return 1 if okay
static int my_init_mng(PluginInstance *This)
{
	mng_retcode rv;
	int err;

	This->mng = mng_initialize((mng_ptr)This,memallocfunc,memfreefunc,NULL);
	//(mng_memalloc)  (mng_memfree)

#ifdef MNGCTRL_CMS
	init_color_management(This);
#endif

	err=0;
	rv=mng_setcb_openstream    (This->mng, callback_openstream   ); if(rv) err++;
	rv=mng_setcb_closestream   (This->mng, callback_closestream  ); if(rv) err++;
	rv=mng_setcb_readdata      (This->mng, callback_readdata     ); if(rv) err++;
	rv=mng_setcb_processheader (This->mng, callback_processheader); if(rv) err++;
	rv=mng_setcb_getcanvasline (This->mng, callback_getcanvasline); if(rv) err++;
	rv=mng_setcb_refresh       (This->mng, callback_refresh      ); if(rv) err++;
	rv=mng_setcb_gettickcount  (This->mng, callback_gettickcount ); if(rv) err++;
	rv=mng_setcb_settimer      (This->mng, callback_settimer     ); if(rv) err++;
	rv=mng_setcb_processtext   (This->mng, callback_processtext  ); if(rv) err++;

#ifdef MNGCTRL_TRACE
	rv=mng_setcb_traceproc     (This->mng, callback_traceproc    ); if(rv) err++;
#endif
	if(err) {
		warn(This,"Error setting libmng callback functions");
		return 0;
	}

	rv= mng_set_suspensionmode (This->mng,MNG_TRUE);
	if(rv) {
		warn(This,"Error setting suspension mode");
		return 0;
	}

	// if the web page author provided a bgcolor, use it
	//if(This->force_bgcolor) {
		rv=mng_set_bgcolor (This->mng, This->bg_r, This->bg_g, This->bg_b);
	//}

#ifdef MNGCTRL_TRACE
	fprintf(tracefile,"initial readdisplay\n");
#endif

	handle_read_error(This, mng_readdisplay(This->mng) );
	return 1;
}

/* Global initialization */
PLGERRCODE PLG_Initialize(HMODULE hInst, HMODULE hInstResource,
		  int arch, int apiversmajor, int apiversminor)
{
	if(!hInst) {
		warn(NULL,"%s error: Cannot load resources",globals.appname);
	}
	globals.hInst           = hInst;
	globals.hInstResource   = hInstResource;
	globals.arch            = arch;
	globals.apiversionmajor = apiversmajor;
	globals.apiversionminor = apiversminor;

	switch(globals.arch) {
	case PLGARCH_NETSCAPE:
		strcpy(globals.appname,"MNGPLG Plug-in");
		break;
	case PLGARCH_ACTIVEX:
		strcpy(globals.appname,"MNG4IE");
		break;
	case PLGARCH_STANDALONE:
		strcpy(globals.appname,"MNGSV");
		break;
	default:
		strcpy(globals.appname,"?");
	}

#ifdef MNGCTRL_TRACE
	tracefile=fopen("c:\\temp\\mngtrace.txt","w");
#endif

#ifndef IDC_HAND
#define IDC_HAND MAKEINTRESOURCE(32649)
#endif
	globals.hcurHand = LoadCursor(NULL,IDC_HAND);
	if(!globals.hcurHand) {
		if(globals.arch==PLGARCH_NETSCAPE)
			globals.hcurHand=LoadCursor(globals.hInstResource,"CURHAND_NS");
		else
			globals.hcurHand=LoadCursor(globals.hInstResource,"CURHAND_IE");
	}

	globals.hfontMsg=CreateFont(-12,0,0,0,FW_DONTCARE,TRUE,0,0,ANSI_CHARSET,
		OUT_DEFAULT_PRECIS,CLIP_DEFAULT_PRECIS,DRAFT_QUALITY,
		VARIABLE_PITCH|FF_SWISS,"Arial");
	return PLGERR_NO_ERROR;
}

/* Global shutdown */
void PLG_Shutdown(void)
{
#ifdef MNGCTRL_TRACE
	if(tracefile) {
		fclose(tracefile);
		tracefile=NULL;
	}
#endif
	if(globals.hfontMsg) DeleteObject((HGDIOBJ)globals.hfontMsg);
	return;
}


static unsigned char gethex(const char *s)
{
	int v[2];
	int i;

	v[0]=v[1]=0; 
	for(i=0;i<2;i++) {
		if(s[i]>='a' && s[i]<='f') v[i]=s[i]-87;
		if(s[i]>='A' && s[i]<='F') v[i]=s[i]-55;
		if(s[i]>='0' && s[i]<='9') v[i]=s[i]-48;
	}
	return (unsigned char)(v[0]*16+v[1]);
}

static void hexcolor2rgb(const char *s, mng_uint16 *r, mng_uint16 *g, mng_uint16 *b)
{
	if(lstrlen(s)!=7) return;
	if(s[0]!='#') return;
	(*r)= gethex(&s[1]); (*r)= (mng_uint16) (((*r)<<8)|(*r));
	(*g)= gethex(&s[3]); (*g)= (mng_uint16) (((*g)<<8)|(*g));
	(*b)= gethex(&s[5]); (*b)= (mng_uint16) (((*b)<<8)|(*b));
}

static void find_window_size(PluginInstance *This)
{
	RECT r;
	if(This->scrolling) {  // make sure scrollbars exist if needed
		ShowScrollBar(This->fhWnd,SB_BOTH,TRUE);
	}
	GetClientRect(This->fhWnd, &r);
	This->windowwidth=r.right;
	This->windowheight=r.bottom;
}

static void set_scrollbars(PluginInstance *This)
{
	SCROLLINFO si;
	int maxpos;

	if(!This->scrolling) return;
	if(!This->fhWnd) return;

	ZeroMemory(&si,sizeof(SCROLLINFO));
	si.cbSize = sizeof(SCROLLINFO);

	// horizontal
	if(This->lpdib) {
		maxpos=This->lpdibinfo->biWidth-This->windowwidth;
		if(maxpos<0) maxpos=0;
		if(This->xscrollpos>maxpos) This->xscrollpos=maxpos;
		if(This->xscrollpos<0) This->xscrollpos=0;
		
		si.fMask = SIF_ALL|SIF_DISABLENOSCROLL;
		si.nMin  = 0;
		si.nMax  = This->lpdibinfo->biWidth -1;
		si.nPage = This->windowwidth;
		si.nPos  = This->xscrollpos;
	}
	else {  // no image to display
		si.fMask = SIF_ALL|SIF_DISABLENOSCROLL;
		si.nMin  = 0;
		si.nMax  = 0;
		si.nPage = 1;
		si.nPos  = 0;
	}
	SetScrollInfo(This->fhWnd,SB_HORZ,&si,TRUE);

	// vertical
	if(This->lpdib) {
		maxpos=This->lpdibinfo->biHeight-This->windowheight;
		if(maxpos<0) maxpos=0;
		if(This->yscrollpos>maxpos) This->yscrollpos=maxpos;
		if(This->yscrollpos<0) This->yscrollpos=0;

		si.fMask = SIF_ALL|SIF_DISABLENOSCROLL;
		si.nMin  = 0;
		si.nMax  = This->lpdibinfo->biHeight -1;
		si.nPage = This->windowheight;
		si.nPos  = This->yscrollpos;
	}
	SetScrollInfo(This->fhWnd,SB_VERT,&si,TRUE);
}


#define SCROLLLINE  40

static void scrollmsg(PluginInstance *This, UINT msg,int code, short int pos)
{
	int page;
	int dx, dy;     // amount of scrolling
	int x_orig, y_orig;

	if(!This->scrolling) return;
	if(!This->lpdib) return;

	x_orig=This->xscrollpos;
	y_orig=This->yscrollpos;

	if(msg==WM_HSCROLL) {
		page=This->windowwidth-15;
		if(page<SCROLLLINE) page=SCROLLLINE;

		switch(code) {
		case SB_LINELEFT: This->xscrollpos-=SCROLLLINE; break;
		case SB_LINERIGHT: This->xscrollpos+=SCROLLLINE; break;
		case SB_PAGELEFT: This->xscrollpos-=page; break;
		case SB_PAGERIGHT: This->xscrollpos+=page; break;
		case SB_LEFT: This->xscrollpos=0; break;
		case SB_RIGHT: This->xscrollpos=This->lpdibinfo->biWidth; break;
		case SB_THUMBTRACK: This->xscrollpos=pos; break;
		default: return;
		}
		set_scrollbars(This);
	}
	else if(msg==WM_VSCROLL) {
		page=This->windowheight-15;
		if(page<SCROLLLINE) page=SCROLLLINE;

		switch(code) {
		case SB_LINEUP: This->yscrollpos-=SCROLLLINE; break;
		case SB_LINEDOWN: This->yscrollpos+=SCROLLLINE; break;
		case SB_PAGEUP: This->yscrollpos-=page; break;
		case SB_PAGEDOWN: This->yscrollpos+=page; break;
		case SB_TOP: This->yscrollpos=0; break;
		case SB_BOTTOM: This->yscrollpos=This->lpdibinfo->biHeight; break;
		case SB_THUMBTRACK: This->yscrollpos=pos; break;
		default: return;
		}
		set_scrollbars(This);
	}

	dx= x_orig - This->xscrollpos;
	dy= y_orig - This->yscrollpos;

	if(dx || dy) {  // if any change
		// GetClientRect(This->fhWnd,&cliprect);
		ScrollWindowEx(This->fhWnd,dx,dy,NULL,NULL /*&cliprect*/,NULL,NULL,SW_INVALIDATE);
	}
}

/* Once-per-instance initialization */
//static PLGERRCODE PLG_New(NPMIMEType pluginType,NPP instance,uint16 mode,
//		int16 argc,char* argn[],char* argv[],NPSavedData* saved)
PLGERRCODE PLG_New(PLGINST instance, int mode,
				   int bgcolorknown, COLORREF bgcolor)
{
	PluginInstance* This;
	//int i;
	int x;
	
	if (instance == NULL) {
		return PLGERR_INVALID_INSTANCE_ERROR;
	}
	instance->pdata = calloc(sizeof(PluginInstance),1);


	This = (PluginInstance*) instance->pdata;
	if (This == NULL) {
	    return PLGERR_OUT_OF_MEMORY_ERROR;
	}

	//This->force_bgcolor=1;
	This->bg_r = This->bg_g = This->bg_b = 0xffff;

	if(bgcolorknown) {
		x=GetRValue(bgcolor); This->bg_r = (mng_uint16) (((mng_uint16)x<<8)|x);
		x=GetGValue(bgcolor); This->bg_g = (mng_uint16) (((mng_uint16)x<<8)|x);
		x=GetBValue(bgcolor); This->bg_b = (mng_uint16) (((mng_uint16)x<<8)|x);
	}

	/* record some info for later lookup */
	//This->fWindow = NULL;
	This->fMode = mode;
	
	This->fhWnd = NULL;
	This->fDefaultWindowProc = NULL;
	This->instance = instance;  /* save the instance id for reverse lookups */
	This->scrolling = (mode==PLGMODE_FULL);
	This->xscrollpos = This->yscrollpos = 0;
	This->windowwidth = This->windowheight = 0;

	This->loadstate = STATE_INIT;
	This->paintedyet = 0;
	This->mng=0;
	This->lpdib=NULL;
	lstrcpy((char*)This->url,"");
	This->frozen=0;
	This->needresume=0;
	This->errorflag=0;
	lstrcpy(This->errormsg,"");

	This->dibsize = This->filesize = 0;

	lstrcpy((char*)This->linkurl,"");
	lstrcpy((char*)This->linktarget,"_self");
	This->islink=0;
	
	This->timer_set=0;
	This->timer2_set=0;
	This->dynamicmng= -1;
	This->mouse_over_mng=0;
	This->mouse_captured=0;

	//This->linkcursor = globals.hcurHandNS;


	return PLGERR_NO_ERROR;
}

// called once for each <embed> tag argument or <param> tag
PLGERRCODE PLG_Param(PLGINST instance,const char*argn, const char*argv)
{
	PluginInstance *This;
	This = (PluginInstance*)instance->pdata;
	if(!This) return PLGERR_OUT_OF_MEMORY_ERROR;

	if(!_stricmp(argn,"bgcolor")) {
		//This->force_bgcolor=1;
		hexcolor2rgb(argv,&This->bg_r,&This->bg_g,&This->bg_b);
	}
	else if(!_stricmp(argn,"href")) {
		lstrcpyn((char*)This->linkurl,argv,MAXLEN_URL);
		This->islink=1;
	}
	else if(!_stricmp(argn,"target")) {
		lstrcpyn((char*)This->linktarget,argv,MAXLEN_TARGET);
	}

	return PLGERR_NO_ERROR;
}

PLGERRCODE PLG_ParamDone(PLGINST instance)
{
	PluginInstance *This;
	This = (PluginInstance*)instance->pdata;

	//This->bkgdbrush=NULL;
	//if(This->force_bgcolor)
		This->bkgdbrush=CreateSolidBrush(RGB(This->bg_r,This->bg_g,This->bg_b));

	return PLGERR_NO_ERROR;
}



static void BeforeDestroyWindow(PluginInstance *This)
{
	if(This->timer_set) {
		KillTimer(This->fhWnd,1);
		This->timer_set=0;
	}
	if(This->timer2_set) {
		KillTimer(This->fhWnd,2);
		This->timer2_set=0;
	}
	if(This->mouse_captured) {
		ReleaseCapture();
		This->mouse_captured=0;
	}

	SetWindowLong(This->fhWnd, GWL_WNDPROC, (LONG)This->fDefaultWindowProc); // unsubclass
	RemoveProp(This->fhWnd,gInstanceLookupString);
	This->fDefaultWindowProc = NULL;
	This->fhWnd = NULL;
}

PLGERRCODE PLG_Destroy(PLGINST instance)
{
	PluginInstance* This;

	if (instance == NULL) return PLGERR_INVALID_INSTANCE_ERROR;
	This = (PluginInstance*) instance->pdata;

	if(!This) return PLGERR_INVALID_INSTANCE_ERROR;

	if(This->mng) {
		This->dynamicmng=0;
		mng_cleanup(&This->mng); 
		This->mng=0;
	}
	if(This->lpdib) {
		free(This->lpdib);
		This->lpdib=NULL;
	}
	if(This->mngdata) {
		free(This->mngdata);
		This->mngdata=NULL;
		This->bytesalloc=0;
	}
	if(This->textdata) {
		free(This->textdata);
		This->textdata=NULL;
	}

	if( This->fhWnd ) { // un-subclass the plugin window
		BeforeDestroyWindow(This);
	}

	if(This->bkgdbrush) DeleteObject((HGDIOBJ)This->bkgdbrush);

	if(This) {
		if(instance->pdata) free(instance->pdata);
		instance->pdata = NULL;
	}

	return PLGERR_NO_ERROR;
}


PLGERRCODE PLG_NewWindow(PLGINST instance, HWND hwnd)
{
	PluginInstance *This;
	PLGERRCODE result;

	This = (PluginInstance*)instance->pdata;
	result = PLGERR_NO_ERROR;
	/* Subclass the new window so that we can begin drawing and
	 * receiving window messages. */
	This->fhWnd = hwnd;
	SetProp(This->fhWnd, gInstanceLookupString, (HANDLE)This);
	This->fDefaultWindowProc = (WNDPROC)SetWindowLong(hwnd,GWL_WNDPROC,(LONG)PluginWindowProc);

	find_window_size(This);

	set_scrollbars(This);

	InvalidateRect(This->fhWnd, NULL, TRUE);
	UpdateWindow(This->fhWnd);

	return result;
}

PLGERRCODE PLG_DestroyWindow(PLGINST instance)
{
	PluginInstance *This;
	This = (PluginInstance*)instance->pdata;

	BeforeDestroyWindow(This);
	return PLGERR_NO_ERROR;
}


// browser is announcing its intent to send data to us
PLGERRCODE PLG_NewStream(PLGINST instance, void*streamid,
						 const char* url, unsigned int streamsize)
{
	PluginInstance* This;

	//if(instance==NULL)
	//	return PLGERR_INVALID_INSTANCE_ERROR;

	This = (PluginInstance*) instance->pdata;
	if(!This)
		return PLGERR_GENERIC_ERROR;

	/* save the URL for later */
	lstrcpyn((char*)This->url,url,MAX_PATH);


	This->libmngpos=0;
	This->bytesloaded=0;
	This->bytesalloc=0;
	This->byteswanted=0;
	This->mngdata=NULL;
	This->textdata=NULL;

	// if we know the total length of the stream in advance 
	// (most of the time we will, hopefully), allocate that amount.
	if(streamsize > 0) {
		This->mngdata = malloc(streamsize);
		This->bytesalloc= streamsize;
	}

	my_init_mng(This);

	This->loadstate=STATE_LOADING;

	//(*stype)=NP_NORMAL;

	return PLGERR_NO_ERROR;
}

#define ALLOC_CHUNK_SIZE 131072

PLGERRCODE PLG_Write(PLGINST instance, void*streamid, int offset, int len, void*buffer)
{
	PluginInstance* This;

#ifdef MNGCTRL_TRACE
	fprintf(tracefile,"PLG_Write offs=%d len=%d\n",offset,len);
#endif

	This = (PluginInstance*) instance->pdata;
	if(!This) return PLGERR_GENERIC_ERROR;
	if(len<1) return PLGERR_NO_ERROR;

	if(offset+len > (int)This->bytesalloc) {  // oops, overflowed our memory buffer
		This->bytesalloc += ALLOC_CHUNK_SIZE;
		if(This->mngdata) {
			This->mngdata=realloc(This->mngdata, This->bytesalloc);
		}
		else {  // first time
			This->mngdata=malloc(This->bytesalloc);
		}
		if(!This->mngdata) {
			warn(This,"Cannot allocate memory for image (%d,%d,%p",offset,len,buffer);
			return PLGERR_OUT_OF_MEMORY_ERROR;
		}
	}

	// now we should have enough room to copy the data to memory

	CopyMemory(&This->mngdata[offset],buffer,len);

	This->bytesloaded = offset+len;

	// now, check if it's time to call mng_read_resume
	if(This->needresume &&
		(This->bytesloaded >= (This->libmngpos + This->byteswanted)) )
	{
		This->needresume=0;
//		handle_read_error(This, mng_read_resume(This->mng) );

#ifdef MNGCTRL_TRACE
	fprintf(tracefile,"PLG_Write display_resume bytesloaded=%d libmngpos=%d byteswanted=%d\n",
		This->bytesloaded,This->libmngpos,This->byteswanted);
#endif

		handle_read_error(This, mng_display_resume(This->mng) );
	}


	//return len; // The number of bytes accepted -- we always accept them all.
	return PLGERR_NO_ERROR;
}

/* DestroyStream gets called after the file has finished loading,
 */
PLGERRCODE PLG_DestroyStream(PLGINST instance, void*streamid, int reason)
{
	PluginInstance* This;
	//if(!instance) return PLGERR_INVALID_INSTANCE_ERROR;

	This = (PluginInstance*) instance->pdata;
//	if(reason==NPRES_DONE) {
	This->filesize = This->bytesloaded;
	This->loadstate = STATE_LOADED;
//	}

	if(reason!=PLGRES_DONE) {
		set_error(This,"Image load failed or was canceled (%d)",(int)reason);
		This->needresume=0;
		if(This->timer_set) { KillTimer(This->fhWnd,1); This->timer_set=0; }
		return PLGERR_NO_ERROR;
	}

#ifdef MNGCTRL_TRACE
	fprintf(tracefile,"PLG_DestroyStream reason=%d needresume=%d\n",reason,This->needresume);
#endif

	if(This->needresume) {
		This->needresume=0;
//		handle_read_error(This, mng_read_resume(This->mng) );
		handle_read_error(This, mng_display_resume(This->mng) );
//		This->needresume=0;
	}

	return PLGERR_NO_ERROR;
}

void PLG_EmbedPrint(PLGINST instance, HDC pdc,
		int x, int y, unsigned int width, unsigned int height)
{
	PluginInstance* This;
	int prevstretchmode;

	This = (PluginInstance*) instance->pdata;

	if(This->loadstate < STATE_VALIDFRAME) return;
	if(!This->lpdib) return;

	prevstretchmode=SetStretchBltMode(pdc,COLORONCOLOR);
	StretchDIBits(pdc,
		x,y,width,height, /* dest coords */
		0,0,This->lpdibinfo->biWidth, This->lpdibinfo->biHeight,   /* source coords */
		This->lpdibbits, (LPBITMAPINFO)This->lpdib,
		DIB_RGB_COLORS,SRCCOPY);
	if(prevstretchmode) SetStretchBltMode(pdc,prevstretchmode);
}

void PLG_FullPrint(PLGINST instance)
{
	return;
}


/**********************************************************************/

/* Try to make a filename from the url. Caller must provide fn[MAX_PATH] buffer.
 * This function attempts to extract a bitmap filename from a URL,
 * but if it doesn't look like it contains an appropriate name,
 * it leaves it blank. */
static void url2filename(char *fn, char *url)
{
	int title,ext,i;

	lstrcpy(fn,"");
	ext=0;   /* position of the file extention */
	title=0; /* position of the base filename */
	for(i=0;url[i];i++) {
		if(url[i]=='.') ext=i+1;
		if(url[i]=='/') title=i+1;
		if(url[i]=='\\') title=i+1;  // handle Microsoft's bogus file: "URLs"
		if(url[i]==':') title=i+1;
		if(url[i]=='=') title=i+1;
	}

	if (!_stricmp(&url[ext],"mng") ||
		!_stricmp(&url[ext],"jng") ||
		!_stricmp(&url[ext],"png") )
	{
		lstrcpyn(fn,&url[title],MAX_PATH);
	}
}

// sanitize string and escape '&'s for use in a menu
static void escapeformenu(unsigned char *s1)
{
	int f, t, len;
	unsigned char s2[200];

	t=0;
	len=lstrlen((char*)s1); if(len>50) len=50;
	for(f=0;f<len;f++) {
		if(s1[f]=='&') {
			s2[t++]='&';
			s2[t++]='&';
		}
		else if(s1[f]<32) {
			s2[t++]='_';
		}
		else {
			s2[t++]=s1[f];
		}
	}
	s2[t]='\0';
	lstrcpy((char*)s1,(char*)s2);
}


static char *get_imagetype_name(mng_imgtype t)
{
	switch(t) {
	case mng_it_mng: return "MNG";
	case mng_it_png: return "PNG";
	case mng_it_jng: return "JNG";
	}
	return "Unknown";
}


/* Write the image to a local file  */
static void SaveImage(PluginInstance *This)
{
	OPENFILENAME ofn;
	char fn[MAX_PATH];
	HANDLE hfile;
	BOOL b;
	mng_imgtype t;
	DWORD byteswritten;

	if(!This->mng || This->loadstate<STATE_LOADED ||
		This->bytesloaded != This->filesize)
	{
		warn(This,"Image not loaded -- can't save");
		return;
	}

	if(lstrlen((char*)This->url)) {
		url2filename(fn,(char*)This->url);
	}
	else {
		lstrcpy(fn,"");
	}

	ZeroMemory(&ofn,sizeof(OPENFILENAME));
	ofn.lStructSize=sizeof(OPENFILENAME);
	ofn.hwndOwner=This->fhWnd;
	ofn.nFilterIndex=1;
	ofn.lpstrTitle="Save Image As...";
	ofn.lpstrFile=fn;
	ofn.nMaxFile=MAX_PATH;
	ofn.Flags=OFN_PATHMUSTEXIST|OFN_HIDEREADONLY|OFN_OVERWRITEPROMPT;

	t=mng_get_sigtype(This->mng);
	if(t==mng_it_png) {
		ofn.lpstrDefExt="png";  // FIXME also give an option of MNG
		ofn.lpstrFilter="PNG (*.png)\0*.png\0\0";
	}
	else if(t==mng_it_jng) {
		ofn.lpstrDefExt="jng";  // FIXME also give an option of MNG
		ofn.lpstrFilter="JNG (*.jng)\0*.jng\0\0";
	}
	else {
		ofn.lpstrFilter="MNG (*.mng)\0*.mng\0\0";
		ofn.lpstrDefExt="mng";
	}

	if(GetSaveFileName(&ofn)) {
		// save to filename: ofn.lpstrFile
		hfile=CreateFile(ofn.lpstrFile,GENERIC_WRITE,FILE_SHARE_READ,
			NULL,CREATE_ALWAYS,FILE_ATTRIBUTE_NORMAL,NULL);
		if(hfile==INVALID_HANDLE_VALUE) {
			warn(This,"Unable to write file");
		}
		else {
			b=WriteFile(hfile, This->mngdata, This->filesize,
				&byteswritten,NULL);
			if(!b || byteswritten != This->filesize) {
				warn(This,"Error writing file");
			}
			CloseHandle(hfile);
		}
	}
}


static void CopyToClipboard(PluginInstance *This,unsigned char *mem,int size,UINT format)
{
	HGLOBAL hClip;
	LPVOID lpClip;

	if(!mem) return;

	if(!OpenClipboard(NULL)) {
		warn(This,"Can't open the clipboard");
		return;
	}

	if(EmptyClipboard()) {
		hClip=GlobalAlloc(GMEM_ZEROINIT|GMEM_MOVEABLE|GMEM_DDESHARE,size);
		lpClip=GlobalLock(hClip);
		if(lpClip) {
			CopyMemory(lpClip,mem,size);
			GlobalUnlock(hClip);
			if(!SetClipboardData(format,hClip)) {
				warn(This,"Can't set clipboard data");
			}
		}
		else {
			warn(This,"Can't allocate memory for clipboard");
		}
	}
	else {
		warn(This,"Can't clear the clipboard");
	}
	CloseClipboard();
}

static void AboutDialog(PluginInstance *This)
{
	DialogBoxParam(globals.hInstResource,"ABOUTDLG",This->fhWnd,(DLGPROC)DlgProcAbout,(LPARAM)This);
}

static void PropDialog(PluginInstance *This)
{
	//if(This->textdata)
	DialogBoxParam(globals.hInstResource,"PROPDLG",This->fhWnd,(DLGPROC)DlgProcProp,(LPARAM)This);
}

static void display_last_error(PluginInstance *This)
{
	if(This->errorflag) {
		warn(This,"%s",This->errormsg);
	}
}

static void DynamicMNG_FireEvent(PluginInstance *This, mng_uint8 eventtype, POINTS pos)
{
	mng_retcode r;
	if(!This->mng) return;
	if(This->dynamicmng ==  0) return;
	if(This->dynamicmng == -1) {
		r=mng_status_dynamic(This->mng);
		if(r==MNG_FALSE) {
			return;
		}
		else {
			This->dynamicmng=1;
		}
	}
	mng_trapevent(This->mng, eventtype, pos.x+This->xscrollpos, pos.y+This->yscrollpos);
}


static void ContextMenu(PluginInstance *This, HWND hwnd)
{
	int cmd;
	HMENU menu;
	POINT pt;
	unsigned char buf[MAX_PATH], buf2[200];

	pt.x=0; pt.y=0;
	GetCursorPos(&pt);

	// create context menu dynamically
	menu=CreatePopupMenu();
	if(This->errorflag) {
		AppendMenu(menu,MF_ENABLED,ID_SHOWERROR,"SHOW ERROR MESSAGE");
		AppendMenu(menu,MF_SEPARATOR,0,NULL);
	}

	AppendMenu(menu,(This->loadstate>=STATE_LOADED?MF_ENABLED:MF_GRAYED),ID_SAVEAS,"Save Image &As...");
	AppendMenu(menu,(This->lpdib?MF_ENABLED:MF_GRAYED),ID_COPYIMAGE,"&Copy Image");
	AppendMenu(menu,MF_ENABLED,ID_COPYURL,"Cop&y Image Location");
	if(This->islink) {
		AppendMenu(menu,MF_ENABLED,ID_COPYLINKLOC,"Copy Link Location");
	}

	if(globals.arch!=PLGARCH_STANDALONE) {
		url2filename((char*)buf,(char*)This->url);
		escapeformenu(buf);
		if(lstrlen((char*)buf)) {
			wsprintf((char*)buf2,"View Image (%s)",buf);
		}
		else {
			wsprintf((char*)buf2,"View Image");
		}
		AppendMenu(menu,MF_ENABLED,ID_VIEWIMAGE,(char*)buf2);
	}

	AppendMenu(menu,MF_SEPARATOR,0,NULL);
	// AppendMenu(menu,(This->mng?MF_ENABLED:MF_GRAYED),ID_STOPANIM,"Stop Animation");


	AppendMenu(menu,(This->mng?MF_ENABLED:MF_GRAYED)|
		(This->frozen?MF_CHECKED:MF_UNCHECKED),ID_FREEZE,"&Freeze Animation");

	// AppendMenu(menu,(This->mng?MF_ENABLED:MF_GRAYED),ID_RESTARTANIM,"Restart Animation");

	AppendMenu(menu,MF_SEPARATOR,0,NULL);

	AppendMenu(menu,MF_ENABLED,ID_PROPERTIES,"Properties...");

	AppendMenu(menu,MF_ENABLED,ID_ABOUT,"About " MNGCTRL_NAME "...");

	cmd=TrackPopupMenuEx(menu, TPM_LEFTALIGN|TPM_TOPALIGN|TPM_NONOTIFY|TPM_RETURNCMD|
		TPM_RIGHTBUTTON,pt.x,pt.y,hwnd,NULL);

	DestroyMenu(menu);

	switch(cmd) {

	case ID_STOPANIM:
		if(This->mng) {
			KillTimer(This->fhWnd,1);
			This->timer_set=0;
			mng_display_freeze(This->mng);
		}
		break;

	case ID_FREEZE:
		This->frozen = !This->frozen;
		if(This->frozen) {
			KillTimer(This->fhWnd,1);
			This->timer_set=0;
			mng_display_freeze(This->mng);
		}
		else {
			handle_read_error(This, mng_display_resume(This->mng) );

		}
		break;

	case ID_RESTARTANIM:
		if(!This->frozen) {
			KillTimer(This->fhWnd,1);
			This->timer_set=0;
			mng_display_freeze(This->mng);
		}
		This->frozen=1;
		mng_display_reset(This->mng);
		This->frozen=0;
		handle_read_error(This, mng_display_resume(This->mng) );
		break;

	case ID_SAVEAS:
		SaveImage(This);
		break;
	case ID_COPYIMAGE:
		if(This->lpdib) {
			CopyToClipboard(This,(unsigned char*)This->lpdib,This->dibsize,CF_DIB);
		}
		else {
			warn(This,"No image to copy");
		}
		break;
	case ID_COPYURL:
		CopyToClipboard(This,This->url,lstrlen((char*)This->url)+1,CF_TEXT);
		break;
	case ID_COPYLINKLOC:
		if(This->islink) {
			CopyToClipboard(This,This->linkurl,lstrlen((char*)This->linkurl)+1,CF_TEXT);
		}
		break;
	case ID_VIEWIMAGE:
		if(lstrlen((char*)This->url)) 
			PLGSERVICE_GetURL(This->instance,(char*)This->url,"_self");
		break;
	case ID_PROPERTIES:
		PropDialog(This);
		break;
	case ID_ABOUT:
		AboutDialog(This);
		break;
	case ID_SHOWERROR:
		display_last_error(This);
		break;
	}
}

/*+++++++++++++++++++++++++++++++++++++++++++++++++
 * PluginWindowProc
 * Handle the Windows window-event loop.
 +++++++++++++++++++++++++++++++++++++++++++++++++*/
static LRESULT CALLBACK PluginWindowProc( HWND hWnd, UINT Msg, WPARAM wParam, LPARAM lParam)
{
	PluginInstance* This;
	HDC hdc;
	RECT rect;
		
	This = (PluginInstance*) GetProp(hWnd, gInstanceLookupString);

	if(!This) return DefWindowProc( hWnd, Msg, wParam, lParam);

	switch(Msg) {

	case WM_ERASEBKGND:
		{
			HBRUSH br;
			hdc= (HDC)wParam;

			if(This->bkgdbrush)
				br=This->bkgdbrush;
			else
				br=GetStockObject(GRAY_BRUSH);

			GetClientRect(hWnd,&rect);
			FillRect(hdc,&rect,br);
			return 1;
		}

	case WM_HSCROLL:
	case WM_VSCROLL:
		scrollmsg(This,Msg,(int)(LOWORD(wParam)),(short int)(HIWORD(wParam)));
		return 0;

	case WM_SIZE:
		find_window_size(This);
		set_scrollbars(This);


		//if(This->errorflag || !This->lpdib) {
		if(globals.arch==PLGARCH_STANDALONE) {
			InvalidateRect(hWnd,NULL,TRUE);
		}
		return 0;


	case WM_CONTEXTMENU: case WM_RBUTTONUP:
		ContextMenu(This, hWnd);
		return 0;


	case WM_SETCURSOR:
		if(LOWORD(lParam)==HTCLIENT) {
			if(This->islink) {
				SetCursor(globals.hcurHand);
				return 1;
			}
		}
		break;

	case WM_LBUTTONDOWN:
		SetCapture(This->fhWnd);
		This->mouse_captured=1;

		if(This->dynamicmng && This->mng && !This->errorflag) {
			DynamicMNG_FireEvent(This,4,MAKEPOINTS(lParam));
		}
		return 0;
		
	case WM_LBUTTONUP:
		{
			RECT rc;
			POINT pt;

			if(This->mouse_captured) {
				ReleaseCapture();
				This->mouse_captured=0;
			}
			if(This->dynamicmng && This->mng && !This->errorflag) {
				DynamicMNG_FireEvent(This,5,MAKEPOINTS(lParam));
			}

			// if mouse is not over image, don't follow links, etc.
			GetWindowRect(This->fhWnd,&rc);
			GetCursorPos(&pt);
			if(!PtInRect(&rc,pt)) return 0;

			if(This->islink) {
				PLGSERVICE_GetURL(This->instance,(char*)This->linkurl,(char*)This->linktarget);
				return 0;
			}
			else if(This->errorflag) {
				display_last_error(This);
			}
		}
		return 0;

	case WM_MOUSEMOVE:

		if(This->dynamicmng && This->mng && This->lpdib && !This->errorflag) {
			POINTS pos;
			int overimage;

			pos=MAKEPOINTS(lParam);
			overimage=0;
			if(pos.x>=0 && pos.x<This->lpdibinfo->biWidth && pos.y>=0 && pos.y<This->lpdibinfo->biHeight) {
				overimage=1;
			}

			if(overimage) {
				if(This->mouse_over_mng) {
					// mouse is still over image: mouse move event
					DynamicMNG_FireEvent(This,2,pos);  // 2=mouse move
				}
				else {
					// mouse wasn't over the image but now it is: mouse-enter event
					DynamicMNG_FireEvent(This,1,pos); // mouse enter
				}
			}
			else { // mouse not now over image
				if(This->mouse_over_mng) { // ... but it used to be
					pos.x=0; pos.y=0;
					DynamicMNG_FireEvent(This,3,pos); // 3=mouse leave
				}
			}

			This->mouse_over_mng=overimage; // remember for next time

			if(This->mouse_over_mng && (This->dynamicmng==1) ) {
#define MOUSE_POLL_INTERVAL  100    // milliseconds
				if(SetTimer(This->fhWnd,2,MOUSE_POLL_INTERVAL,NULL)) {
					This->timer2_set=1;
				}
			}
		}
		return 0;

	case WM_PAINT:
		{
			PAINTSTRUCT paintStruct;
			HDC hdc;
			RECT rect2;

			hdc = BeginPaint( hWnd, &paintStruct );
			SetWindowOrgEx(hdc,This->xscrollpos,This->yscrollpos,NULL);

			GetClientRect(hWnd,&rect);
			if(This) {
				if(This->errorflag || !This->lpdib) {
					SelectObject(hdc,globals.hfontMsg);
					Rectangle(hdc,rect.left,rect.top,rect.right,rect.bottom);
					rect2.left=rect.left+2;
					rect2.top=rect.top+2;
					rect2.right=rect.right-2;
					rect2.bottom=rect.bottom-2;
					if(This->errorflag) {
						DrawText(hdc,MNGCTRL_NAME " ERROR!",-1,&rect2,DT_LEFT|DT_WORDBREAK);
					}
					else {
						if(This->loadstate>=STATE_LOADING) {
							DrawText(hdc,"MNG image loading...",-1,&rect2,DT_LEFT|DT_WORDBREAK);
						}
						else {
							DrawText(hdc,MNGCTRL_NAME,-1,&rect2,DT_LEFT|DT_WORDBREAK);
						}
					}
				}
				else if(This->lpdib) {
					StretchDIBits(hdc,
						0,0,This->lpdibinfo->biWidth,This->lpdibinfo->biHeight,
						0,0,This->lpdibinfo->biWidth,This->lpdibinfo->biHeight,
						&((BYTE*)(This->lpdib))[sizeof(BITMAPINFOHEADER)],
						(LPBITMAPINFO)This->lpdib,DIB_RGB_COLORS,SRCCOPY);
				}
			}


			EndPaint( hWnd, &paintStruct );
		}
		return 0;

	case WM_TIMER:
		switch(wParam) {
		case 1: // the main animation timer
			KillTimer(hWnd,1);
			This->timer_set=0;

#ifdef MNGCTRL_TRACE
			fprintf(tracefile,"WM_TIMER display_resume bytesloaded=%d\n",This->bytesloaded);
#endif

			if(This->mng) {
				if(!This->needresume) {
					handle_read_error(This, mng_display_resume(This->mng) );
				}
			}
			return 0;

		case 2: // timer for polling mouse position
			{
				RECT rc;
				POINT pt;

				GetWindowRect(hWnd,&rc);
				GetCursorPos(&pt);
				if(!PtInRect(&rc,pt)) {
					POINTS pos;
					KillTimer(hWnd,2);
					This->timer2_set=0;
					pos.x=0; pos.y=0;
					DynamicMNG_FireEvent(This,3,pos); // 3=mouse leave
					This->mouse_over_mng=0;
				}
			}
			return 0;
		}
		break;

	}

	/* Forward unprocessed messages on to their original destination
	 * (the window proc we replaced) */
	return This->fDefaultWindowProc(hWnd, Msg, wParam, lParam);
}

static LRESULT CALLBACK DlgProcProp(HWND hWnd, UINT Msg, WPARAM wParam, LPARAM lParam)
{
	char buf[4096],buf2[1024];

	switch(Msg) {
	case WM_INITDIALOG:
		{
			DWORD tabs[1];

			PluginInstance *This=(PluginInstance*)lParam;

			tabs[0]= 60;
			SendDlgItemMessage(hWnd,IDC_IMGINFO,EM_SETTABSTOPS,(WPARAM)1,(LPARAM)tabs);

			wsprintf(buf,"URL:\t%s\r\n",This->url);

			if(This->lpdib) {
				wsprintf(buf2,"Dimensions:\t%d x %d\r\n",This->lpdibinfo->biWidth,
					This->lpdibinfo->biHeight);
				lstrcat(buf,buf2);
			}

			if(This->lpdib && This->fMode==PLGMODE_EMBED) {
				wsprintf(buf2,"Window:\t%d x %d\r\n",This->windowwidth,
					This->windowheight);
				lstrcat(buf,buf2);
			}

			if(This->filesize) {
				wsprintf(buf2,"File size:\t%u bytes\r\n",This->filesize);
				lstrcat(buf,buf2);
			}

#ifdef _DEBUG
			if(This->mngdata && This->lpdib && This->bytesalloc && This->dibsize) {
				// note this doesn't include memory used by libmng
				wsprintf(buf2,"Memory used:\t%u bytes\r\n",
					This->bytesalloc + This->dibsize);
				lstrcat(buf,buf2);
			}
#endif

			if(This->islink) {
				wsprintf(buf2,"Link to:\t%s\r\n",This->linkurl);
				lstrcat(buf,buf2);
				if(strcmp((char*)This->linktarget,"_self")) {
					wsprintf(buf2,"Link target:\t%s\r\n",This->linktarget);
					lstrcat(buf,buf2);
				}
			}

			if(This->loadstate >= STATE_VALIDFRAME) {
				wsprintf(buf2,"Signature:\t%s\r\n",get_imagetype_name(mng_get_sigtype(This->mng)));
				lstrcat(buf,buf2);
				wsprintf(buf2,"Image type:\t%s\r\n",get_imagetype_name(mng_get_imagetype(This->mng)));
				lstrcat(buf,buf2);
				wsprintf(buf2,"Simplicity:\t0x%08x\r\n",mng_get_simplicity(This->mng));
				lstrcat(buf,buf2);
				wsprintf(buf2,"Frame count:\t%u\r\n",mng_get_framecount(This->mng));
				lstrcat(buf,buf2);
				wsprintf(buf2,"Layer count:\t%u\r\n",mng_get_layercount(This->mng));
				lstrcat(buf,buf2);
				wsprintf(buf2,"Play time:\t%u\r\n",mng_get_playtime(This->mng));
				lstrcat(buf,buf2);
			}

			SetDlgItemText(hWnd,IDC_IMGINFO,buf);

			if(This->textdata)
				SetDlgItemText(hWnd,IDC_MNGTEXT,This->textdata);
		}
		return(TRUE);
	case WM_CLOSE:
		EndDialog(hWnd,0);
		return(TRUE);
	case WM_COMMAND:
		switch(wParam) {
		case IDOK:
		case IDCANCEL:
			EndDialog(hWnd,0);
			return(TRUE);
		}
	}
	return(FALSE);
}

static LRESULT CALLBACK DlgProcAbout(HWND hWnd, UINT Msg, WPARAM wParam, LPARAM lParam)
{
	char buf[4096],buf2[1024],buf3[300];

	switch(Msg) {
	case WM_INITDIALOG:
		{
			//DWORD tabs[1];

			//PluginInstance *This=(PluginInstance*)lParam;

			//tabs[0]= 60;
			//SendDlgItemMessage(hWnd,IDC_IMGINFO,EM_SETTABSTOPS,(WPARAM)1,(LPARAM)tabs);

			wsprintf(buf,"%s, Version %s\r\n%s"
#ifdef _DEBUG
				" DEBUG BUILD"
#endif
				"\r\nCopyright (C) 2000-2004 by Jason Summers\r\n\r\n",
				globals.appname,
				MNGCTRLVERS,__DATE__);

			wsprintf(buf2,"Based on libmng by Gerard Juyn.\r\n");
			lstrcat(buf,buf2);

			wsprintf(buf2,"libmng version: %s\r\n\r\n",mng_version_text());
			lstrcat(buf,buf2);

			wsprintf(buf2,"Uses the zlib compression library.\r\n");
			lstrcat(buf,buf2);
			wsprintf(buf2,"zlib version: %s\r\n\r\n",zlibVersion());
			lstrcat(buf,buf2);

			wsprintf(buf2,"This software is based in part on the work of the "
				"Independent JPEG Group.\r\n");
			lstrcat(buf,buf2);
			// This really only gives the version of the libjpeg header used when
			// compiling this plugin, but I don't know how to query libjpeg for its
			// version.
			wsprintf(buf2,"IJG JPEG library version: %s\r\n%s\r\n\r\n",JVERSION,JCOPYRIGHT);
			lstrcat(buf,buf2);

#ifdef MNGCTRL_CMS
			wsprintf(buf2,"Uses the Little CMS color management library by Marti Maria.\r\n"
			    "Copyright (C) 1998-2003 Marti Maria\r\n");
			lstrcat(buf,buf2);
#ifdef LCMS_VERSION
			wsprintf(buf2,"lcms version: %d.%02d\r\n",LCMS_VERSION/100,LCMS_VERSION%100);
			lstrcat(buf,buf2);
#endif
			lstrcat(buf,"\r\n");
#endif

			if(GetModuleFileName(globals.hInst,buf3,260)) {
				wsprintf(buf2,"Location: %s\r\n",buf3);
				lstrcat(buf,buf2);
			}

			SetDlgItemText(hWnd,IDC_PRGINFO,buf);

		}
		return(TRUE);
	case WM_CLOSE:
		EndDialog(hWnd,0);
		return(TRUE);
	case WM_COMMAND:
		switch(wParam) {
		case IDOK:
		case IDCANCEL:
			EndDialog(hWnd,0);
			return(TRUE);
		}
	}
	return(FALSE);
}


void script_Animate(PLGINST instance, long x)
{
	PluginInstance *This;
	This = (PluginInstance*)instance->pdata;
	if(!This) return;

	if(x) {  // unfreeze
		if(!This->frozen) return;
		This->frozen=0;
		handle_read_error(This,mng_display_resume(This->mng));
	}
	else { // freeze
		if(This->frozen) return;
		This->frozen=1;
		if(This->timer_set) {
			KillTimer(This->fhWnd,1);
			This->timer_set=0;
		}
		mng_display_freeze(This->mng);
	}
}

int script_GetVersion(PLGINST instance)
{
	return MNGCTRLVERS_INT;
}

int script_GetAnimateState(PLGINST instance)
{
	PluginInstance *This;
	This = (PluginInstance*)instance->pdata;
	if(!This) return 0;
	return !This->frozen;
}

void script_AboutBox(PLGINST instance)
{
	PluginInstance *This;
	This = (PluginInstance*)instance->pdata;
	if(!This) return;
	AboutDialog(This);
}

void script_ImageProperties(PLGINST instance)
{
	PluginInstance *This;
	This = (PluginInstance*)instance->pdata;
	if(!This) return;
	PropDialog(This);
}

void script_CopyImage(PLGINST instance)
{
	PluginInstance *This;
	This = (PluginInstance*)instance->pdata;
	if(!This) return;
	if(This->lpdib) {
		CopyToClipboard(This,(unsigned char*)This->lpdib,This->dibsize,CF_DIB);
	}
}

