// mngctrl.h

#ifndef INC_MNGCTRL_H
#define INC_MNGCTRL_H

#define MNGCTRLVERS    "1.0.7"
#define MNGCTRLVERS_INT   1000701


// ---------------
// If you get an error about INT32 in jmorecfg.h, try modifying
// jmorecfg.h by changing:
//   typedef long INT32;
// to:
//   #ifdef JPEG_INTERNALS
//   typedef long INT32;
//   #endif
//   -----------
// If you get errors in lcms.h, try removing or renaming each
// occurrence of the FAR macro.
//   -----------
#define MNG_ERROR_TELLTALE
#define MNG_FULL_CMS
#define MNG_SUPPORT_DYNAMICMNG
#include <libmng.h>  // needed for mng_handle, etc.


#ifdef __cplusplus
#define BEGIN_C_DEFS  extern "C" {
#define END_C_DEFS    }
#else
#define BEGIN_C_DEFS
#define END_C_DEFS
#endif


BEGIN_C_DEFS


#define PLGERR_NO_ERROR                    0
#define PLGERR_GENERIC_ERROR               1
#define PLGERR_INVALID_INSTANCE_ERROR      2
#define PLGERR_INVALID_FUNCTABLE_ERROR     3
#define PLGERR_MODULE_LOAD_FAILED_ERROR    4
#define PLGERR_OUT_OF_MEMORY_ERROR         5
#define PLGERR_INVALID_PLUGIN_ERROR        6
#define PLGERR_INVALID_PLUGIN_DIR_ERROR    7
#define PLGERR_INCOMPATIBLE_VERSION_ERROR  8
#define PLGERR_INVALID_PARAM               9
#define PLGERR_INVALID_URL                 10
#define PLGERR_FILE_NOT_FOUND              11
#define PLGERR_NO_DATA                     12
#define PLGERR_STREAM_NOT_SEEKABLE         13

#define PLGMODE_EMBED  1
#define PLGMODE_FULL   2

#define PLGRES_DONE         0
#define PLGRES_NETWORK_ERR  1
#define PLGRES_USER_BREAK   2

#define PLGARCH_NETSCAPE    1
#define PLGARCH_ACTIVEX     2
#define PLGARCH_STANDALONE  3


typedef unsigned short PLGERRCODE;    // NPError

typedef struct container_struct {
	int reserved;
	void *pdata;

	// private stuff:
	void *npp;
	void *ctl; // activex object
	HWND hwnd;

} PLGINST_t;
typedef PLGINST_t* PLGINST;

struct globals_struct {
	HINSTANCE hInst;
	HINSTANCE hInstResource;
	//HCURSOR hcurHandNS;
	//HCURSOR hcurHandIE;
	HCURSOR hcurHand;
	HFONT hfontMsg;
	int arch;
	int apiversionmajor, apiversionminor;
	char appname[32];
};

/* instance-specific data */
typedef struct plugininst_struct
{
	//NPWindow*  fWindow;
	int        fMode;
	HWND       fhWnd;
	WNDPROC    fDefaultWindowProc;
	PLGINST    instance;

	mng_handle mng;

#define STATE_INIT        0
#define STATE_LOADING     1  // stream opened
#define STATE_VALIDFRAME  2  // at least one frame has been displayed
#define STATE_LOADED      3  // image loaded; stream closed


#define MAXLEN_TEXT   5000

#define MAXLEN_URL     300
#define MAXLEN_TARGET  100

	// I think I'm not doing this very well. Probably there really needs to be
	// two state variables, one for loading from the network, and one for
	// the libmng processing. (or use libmng's new getstate API?)
	int loadstate;

	int paintedyet;

	int scrolling;    // allow scrolling of the image?
	int xscrollpos, yscrollpos;
	int windowwidth, windowheight; // client size of current window

	int diblinesize;
	DWORD dibsize;
	DWORD filesize;
	DWORD libmngpos;  // count of bytes that have been sent to libmng
	DWORD byteswanted;   // libmng asked for this many more bytes (add to libmngpos)

	unsigned char *mngdata;    // stores the MNG file in memory
	DWORD bytesloaded;  // 
	DWORD bytesalloc;   // size of mngdata
	int needresume;   // if previous mng_readdisplay call returned NEEDMOREDATA

	char *textdata;

	int errorflag;    // set if an error occurs that prevents displaying the image
	char errormsg[256];

	unsigned char *lpdib;    // pointer to header section of dib
	unsigned char *lpdibbits;   // pointer to "bits" section of dib (follows the header)
	LPBITMAPINFOHEADER lpdibinfo;  // alias for lpdib
	int frozen;
	int timer_set;
	int timer2_set;
	int dynamicmng;
	int mouse_over_mng;
	int mouse_captured;

	//int force_bgcolor;
	mng_uint16 bg_r,bg_g,bg_b;  // background color

	unsigned char url[MAX_PATH];  // the url of the stream

	int islink;
	HCURSOR linkcursor;
	HBRUSH bkgdbrush;
	unsigned char linkurl[MAXLEN_URL];
	unsigned char linktarget[MAXLEN_TARGET];

} PluginInstance;



PLGERRCODE PLG_Initialize(HMODULE hInst,HMODULE hInstResource,
						  int arch, int apiversmajor, int apiversionminor);
void PLG_Shutdown(void);

PLGERRCODE PLG_New(PLGINST instance, int mode,
				   int bgcolorknown, COLORREF bgcolor);
PLGERRCODE PLG_Param(PLGINST instance,const char*argn, const char*argv);
PLGERRCODE PLG_ParamDone(PLGINST instance);
PLGERRCODE PLG_Destroy(PLGINST instance);

PLGERRCODE PLG_NewWindow(PLGINST instance, HWND hwnd);
PLGERRCODE PLG_DestroyWindow(PLGINST instance);

PLGERRCODE PLG_NewStream(PLGINST instance, void*streamid,
						 const char* url, unsigned int streamsize);
PLGERRCODE PLG_Write(PLGINST instance, void*streamid, int offset, int len, void*buffer);
PLGERRCODE PLG_DestroyStream(PLGINST instance, void*streamid, int reason);

void PLG_EmbedPrint(PLGINST instance, HDC pdc,
		int x, int y, unsigned int width, unsigned int height);
void PLG_FullPrint(PLGINST instance);

const char* PLGSERVICE_UserAgent(PLGINST plginst);
PLGERRCODE PLGSERVICE_GetURL(PLGINST plginst, const char *url, const char *target);


int script_GetVersion(PLGINST instance);
void script_Animate(PLGINST instance, long x);
int script_GetAnimateState(PLGINST instance);
void script_AboutBox(PLGINST instance);
void script_ImageProperties(PLGINST instance);
void script_CopyImage(PLGINST instance);


END_C_DEFS

#endif // INC_MNGCTRL_H
